package cn.com.qmth.examcloud.starters.face.verify.model.baidu;

import cn.com.qmth.examcloud.starters.face.verify.FaceVerifyProperties;
import cn.com.qmth.examcloud.starters.face.verify.common.Constants;
import cn.com.qmth.examcloud.starters.face.verify.common.FaceVerifyException;
import cn.com.qmth.examcloud.starters.face.verify.common.HttpClientBuilder;
import cn.com.qmth.examcloud.starters.face.verify.common.JsonHelper;
import cn.com.qmth.examcloud.starters.face.verify.model.FaceResult;
import com.fasterxml.jackson.annotation.JsonInclude.Include;
import okhttp3.*;
import org.apache.commons.lang3.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.IOException;
import java.util.List;

public class BaiduApiHelper {

    private final static Logger log = LoggerFactory.getLogger(BaiduApiHelper.class);

    private static double expectFaceLivenessScore = 0.39d;

    public static BaiduSession getAccessToken(String apiKey, String apiSecret) {
        if (StringUtils.isBlank(apiKey)) {
            throw new IllegalArgumentException("[BAIDU] apiKey must be not empty.");
        }

        if (StringUtils.isBlank(apiSecret)) {
            throw new IllegalArgumentException("[BAIDU] apiSecret must be not empty.");
        }

        String requestUrl = String.format("%s?grant_type=client_credentials&client_id=%s&client_secret=%s", Constants.BAIDU_TOKEN_URL, apiKey, apiSecret);
        Request.Builder request = new Request.Builder().url(requestUrl).get();

        try (Response response = HttpClientBuilder.getClient().newCall(request.build()).execute();
             ResponseBody body = response.body();) {
            String bodyStr = body != null ? body.string() : null;

            if (response.isSuccessful()) {
                BaiduSession session = new JsonHelper().parseJson(bodyStr, BaiduSession.class);
                if (session != null) {
                    return session;
                }
            }

            log.warn("[BAIDU] url:{} response:{} body:{}", requestUrl, response.code(), bodyStr);
            throw new FaceVerifyException("[BAIDU] 获取AccessToken失败！");
        } catch (IOException e) {
            log.error("[BAIDU] url:{} {}:{}", requestUrl, e.getClass().getSimpleName(), e.getMessage());
            throw new FaceVerifyException("[BAIDU] 获取AccessToken异常！");
        }
    }

    public static FaceResult faceVerify(FaceVerifyProperties properties, String jsonParams) {
        if (StringUtils.isBlank(jsonParams)) {
            throw new IllegalArgumentException("[BAIDU] api params must be not empty.");
        }

        BaiduSession session = BaiduApiHelper.getAccessToken(properties.getBaiduKey(), properties.getBaiduSecret());
        String accessToken = session.getAccess_token();

        String requestUrl = String.format("%s?access_token=%s", Constants.BAIDU_FACE_VERIFY_URL, accessToken);
        RequestBody formBody = FormBody.create(MediaType.parse(Constants.JSON_CONTENT_TYPE), jsonParams);
        jsonParams = null;// 清理大对象内存

        BaiduResponse response = callApi(requestUrl, formBody);
        return parseFaceVerifyResult(response);
    }

    public static FaceResult faceVerifyUsePrivateApi(FaceVerifyProperties properties, String jsonParams) {
        if (StringUtils.isBlank(jsonParams)) {
            throw new IllegalArgumentException("[BAIDU] api params must be not empty.");
        }

        String requestUrl = String.format("%s?appid=%s", properties.getPrivateBaiduUrlPrefix() + Constants.PRIVATE_BAIDU_FACE_VERIFY_URL, properties.getPrivateBaiduAppId());
        RequestBody formBody = FormBody.create(MediaType.parse(Constants.JSON_CONTENT_TYPE), jsonParams);
        jsonParams = null;// 清理大对象内存

        BaiduResponse response = callApi(requestUrl, formBody);
        return parseFaceVerifyResult(response);
    }

    public static FaceResult faceDetect(FaceVerifyProperties properties, String jsonParams) {
        if (StringUtils.isBlank(jsonParams)) {
            throw new IllegalArgumentException("[BAIDU] api params must be not empty.");
        }

        BaiduSession session = BaiduApiHelper.getAccessToken(properties.getBaiduKey(), properties.getBaiduSecret());
        String accessToken = session.getAccess_token();

        String requestUrl = String.format("%s?access_token=%s", Constants.BAIDU_FACE_DETECT_URL, accessToken);
        RequestBody formBody = FormBody.create(MediaType.parse(Constants.JSON_CONTENT_TYPE), jsonParams);
        jsonParams = null;// 清理大对象内存

        BaiduResponse response = callApi(requestUrl, formBody);
        return parseFaceDetectResult(response);
    }

    public static FaceResult faceDetectUsePrivateApi(FaceVerifyProperties properties, String jsonParams) {
        if (StringUtils.isBlank(jsonParams)) {
            throw new IllegalArgumentException("[BAIDU] api params must be not empty.");
        }

        String requestUrl = String.format("%s?appid=%s", properties.getPrivateBaiduUrlPrefix() + Constants.PRIVATE_BAIDU_FACE_DETECT_URL, properties.getPrivateBaiduAppId());
        RequestBody formBody = FormBody.create(MediaType.parse(Constants.JSON_CONTENT_TYPE), jsonParams);
        jsonParams = null;// 清理大对象内存

        BaiduResponse response = callApi(requestUrl, formBody);
        return parseFaceDetectResult(response);
    }

    public static FaceResult faceCompare(FaceVerifyProperties properties, String jsonParams, Double expectFaceCompareScore) {
        if (StringUtils.isBlank(jsonParams)) {
            throw new IllegalArgumentException("[BAIDU] api params must be not empty.");
        }

        BaiduSession session = BaiduApiHelper.getAccessToken(properties.getBaiduKey(), properties.getBaiduSecret());
        String accessToken = session.getAccess_token();

        String requestUrl = String.format("%s?access_token=%s", Constants.BAIDU_FACE_COMPARE_URL, accessToken);
        RequestBody formBody = FormBody.create(MediaType.parse(Constants.JSON_CONTENT_TYPE), jsonParams);
        jsonParams = null;// 清理大对象内存

        BaiduResponse response = callApi(requestUrl, formBody);
        return parseFaceCompareResult(response, expectFaceCompareScore);
    }

    public static FaceResult faceCompareUsePrivateApi(FaceVerifyProperties properties, String jsonParams, Double expectFaceCompareScore) {
        if (StringUtils.isBlank(jsonParams)) {
            throw new IllegalArgumentException("[BAIDU] api params must be not empty.");
        }

        String requestUrl = String.format("%s?appid=%s", properties.getPrivateBaiduUrlPrefix() + Constants.PRIVATE_BAIDU_FACE_COMPARE_URL, properties.getPrivateBaiduAppId());
        RequestBody formBody = FormBody.create(MediaType.parse(Constants.JSON_CONTENT_TYPE), jsonParams);
        jsonParams = null;// 清理大对象内存

        BaiduResponse response = callApi(requestUrl, formBody);
        return parseFaceCompareResult(response, expectFaceCompareScore);
    }

    private static FaceResult parseFaceVerifyResult(BaiduResponse response) {
        FaceResult result = new FaceResult();
        result.setFacePass(false);
        result.setFaceScore(0d);
        result.setFaceNum(0);
        result.setJsonResult(new JsonHelper().style(Include.NON_NULL).toJson(response));

        int errorCode = response.getError_code();
        if (errorCode != 0) {
            result.setApiNeedRetry(true);// 默认值
            result.setError(errorCode + " - " + response.getError_msg());

            if (errorCode == 222202 || errorCode == 222203 || errorCode == 222209 || errorCode == 222304) {
                // 222202 图片中没有人脸
                // 222203 无法解析人脸
                // 222209 face token不存在
                // 222304 图片尺寸太大

                result.setApiNeedRetry(false);
            }

            // 请求并发超限(约20QPS，可按需调整)
            if (errorCode == 18) {
                log.debug("baidu api qps limit...");
            }

            return result;
        }

        BaiduResult data = response.getResult();
        if (data == null) {
            result.setError(result.getJsonResult());
            return result;
        }

        double faceLiveness = data.getFace_liveness() != null ? data.getFace_liveness() : 0d;
        result.setFaceScore(faceLiveness);

        // 注：该接口的faceNum值没多大参考意义！照片中包含多张人脸时，默认取人脸面积最大的人脸，返回face_list的元素数量总不会大于1。
        int faceNum = data.getFace_list() != null ? data.getFace_list().size() : 0;
        result.setFaceNum(faceNum);

        // 活体分数值，推荐阈值0.393241
        if (faceNum == 1 && faceLiveness >= expectFaceLivenessScore) {
            result.setFacePass(true);
        } else {
            result.setError("人脸数量:" + faceNum + "; 真实性得分:" + faceLiveness + ",阈值:" + expectFaceLivenessScore);
        }

        return result;
    }

    private static FaceResult parseFaceDetectResult(BaiduResponse response) {
        FaceResult result = new FaceResult();
        result.setFacePass(false);
        result.setFaceScore(0d);
        result.setFaceNum(0);
        result.setJsonResult(new JsonHelper().style(Include.NON_NULL).toJson(response));

        int errorCode = response.getError_code();
        if (errorCode != 0) {
            result.setApiNeedRetry(true);// 默认值
            result.setError(errorCode + " - " + response.getError_msg());

            if (errorCode == 222202 || errorCode == 222203 || errorCode == 222209 || errorCode == 222304) {
                // 222202 图片中没有人脸
                // 222203 无法解析人脸
                // 222209 face token不存在
                // 222304 图片尺寸太大

                result.setApiNeedRetry(false);
            }

            // 请求并发超限(约20QPS，可按需调整)
            if (errorCode == 18) {
                log.debug("baidu api qps limit...");
            }

            return result;
        }

        BaiduResult data = response.getResult();
        if (data == null) {
            result.setError(result.getJsonResult());
            return result;
        }

        double faceLiveness = 0d;
        List<BaiduFace> faces = data.getFace_list();
        if (faces != null && !faces.isEmpty()) {
            BaiduFace face = faces.get(0);
            if (face.getLiveness() != null && face.getLiveness().getLivemapscore() != null) {
                faceLiveness = face.getLiveness().getLivemapscore();
            }
        }
        result.setFaceScore(faceLiveness);

        // 注：照片中包含多张人脸时，优先取人脸面积最大的人脸，返回的face_list的最大数量由接口请求参数“max_face_num（默认值1）”的值限定。
        int faceNum = faces != null ? faces.size() : 0;
        result.setFaceNum(faceNum);

        // 活体分数值，推荐阈值0.393241
        if (faceNum == 1 && faceLiveness >= expectFaceLivenessScore) {
            result.setFacePass(true);
        } else {
            result.setError("人脸数量:" + faceNum + "; 人脸得分:" + faceLiveness + ",阈值:" + expectFaceLivenessScore);
        }

        return result;
    }

    private static FaceResult parseFaceCompareResult(BaiduResponse response, Double expectFaceCompareScore) {
        if (expectFaceCompareScore == null) {
            throw new IllegalArgumentException("[BAIDU] expectFaceCompareScore must be not null.");
        }

        FaceResult result = new FaceResult();
        result.setFacePass(false);
        result.setFaceScore(0d);
        result.setFaceNum(0);
        result.setJsonResult(new JsonHelper().style(Include.NON_NULL).toJson(response));

        int errorCode = response.getError_code();
        if (errorCode != 0) {
            result.setApiNeedRetry(true);// 默认值
            result.setError(errorCode + " - " + response.getError_msg());

            if (errorCode == 222202 || errorCode == 222203 || errorCode == 222209 || errorCode == 222304) {
                // 222202 图片中没有人脸
                // 222203 无法解析人脸
                // 222209 face token不存在
                // 222304 图片尺寸太大

                result.setApiNeedRetry(false);
            }

            // 请求并发超限(约20QPS，可按需调整)
            if (errorCode == 18) {
                log.debug("baidu api qps limit...");
            }

            return result;
        }

        BaiduResult data = response.getResult();
        if (data == null) {
            result.setError(result.getJsonResult());
            return result;
        }

        double score = data.getScore() != null ? data.getScore() : 0d;
        result.setFaceScore(score);

        // 人脸相似度得分，推荐阈值80
        if (score >= expectFaceCompareScore) {
            result.setFacePass(true);
        } else {
            result.setError("人脸相似度得分:" + score + ",阈值:" + expectFaceCompareScore);
        }

        return result;
    }

    private static BaiduResponse callApi(String requestUrl, RequestBody formBody) {
        long start = System.currentTimeMillis();
        Request.Builder request = new Request.Builder().url(requestUrl).post(formBody);
        try (Response response = HttpClientBuilder.getClient().newCall(request.build()).execute();
             ResponseBody body = response.body();) {
            String bodyStr = body != null ? body.string() : null;

            long cost = System.currentTimeMillis() - start;
            // System.out.println("response:" + response.code() + " cost:" + cost + "ms body:" + bodyStr);

            if (response.isSuccessful()) {
                BaiduResponse result = new JsonHelper().parseJson(bodyStr, BaiduResponse.class);
                if (result != null) {
                    if (log.isDebugEnabled()) {
                        log.debug("[BAIDU] url:{} response:{} cost:{}ms body:{}", requestUrl, response.code(), cost, bodyStr);
                    }
                    return result;
                }
            }

            log.warn("[BAIDU] url:{} response:{} cost:{}ms body:{}", requestUrl, response.code(), cost, bodyStr);
        } catch (IOException e) {
            log.error("[BAIDU] url:{} {}:{}", requestUrl, e.getClass().getSimpleName(), e.getMessage());
        }

        throw new FaceVerifyException("人脸接口调用异常！");
    }

}
