package cn.com.qmth.examcloud.starters.face.verify.service.impl;

import cn.com.qmth.examcloud.starters.face.verify.FaceVerifyProperties;
import cn.com.qmth.examcloud.starters.face.verify.common.CommonUtils;
import cn.com.qmth.examcloud.starters.face.verify.common.JsonHelper;
import cn.com.qmth.examcloud.starters.face.verify.model.FaceResult;
import cn.com.qmth.examcloud.starters.face.verify.model.baidu.BaiduApiHelper;
import cn.com.qmth.examcloud.starters.face.verify.model.faceplus.FacePlusApiHelper;
import cn.com.qmth.examcloud.starters.face.verify.model.param.FaceApiType;
import cn.com.qmth.examcloud.starters.face.verify.model.param.FaceParam;
import cn.com.qmth.examcloud.starters.face.verify.model.param.ImageParm;
import cn.com.qmth.examcloud.starters.face.verify.model.param.ImageParmType;
import cn.com.qmth.examcloud.starters.face.verify.service.FaceVerifyService;
import org.apache.commons.lang3.ArrayUtils;
import org.apache.commons.lang3.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.File;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

public class FaceVerifyServiceImpl implements FaceVerifyService {

    private final static Logger log = LoggerFactory.getLogger(FaceVerifyServiceImpl.class);

    private FaceVerifyProperties properties;

    /**
     * 人脸真实性检测
     */
    @Override
    public FaceResult faceVerify(FaceParam faceParam) {
        if (ArrayUtils.getLength(faceParam.getImages()) < 1) {
            throw new IllegalArgumentException("至少传一个图片参数！");
        }

        if (FaceApiType.FACE_PLUS == faceParam.getApiType()) {
            throw new IllegalArgumentException("FacePlus暂不支持人脸真实性检测！");
        }

        // 百度私有化API：未指定参数值时采用配置值；明确指定时则采用参数值
        boolean usePrivateBaiduApi = faceParam.getApiType() == null ? properties.getDefaultPrivateBaiduApi() :
                FaceApiType.PRIVATE_BAIDU_API == faceParam.getApiType();

        Map<String, String> imageData = this.buildImageParamForBaidu(faceParam.getImages()[0], usePrivateBaiduApi);
        imageData.put("face_field", "spoofing,face_type,gender,age");
        List<Map<String, String>> images = new ArrayList<>();
        images.add(imageData);
        String params = new JsonHelper().toJson(images);

        // 清理大对象内存
        images.clear();
        imageData.clear();
        faceParam = null;

        if (usePrivateBaiduApi) {
            return BaiduApiHelper.faceVerifyUsePrivateApi(properties, params);
        }
        return BaiduApiHelper.faceVerify(properties, params);
    }

    /**
     * 人脸识别
     */
    @Override
    public FaceResult faceDetect(FaceParam faceParam) {
        if (ArrayUtils.getLength(faceParam.getImages()) < 1) {
            throw new IllegalArgumentException("至少传一个图片参数！");
        }

        if (FaceApiType.FACE_PLUS == faceParam.getApiType()) {
            Map<String, String> params = new HashMap<>();
            this.buildImageParamForFacePlus(params, faceParam.getImages()[0], null);

            return FacePlusApiHelper.faceDetect(properties, params);
        }

        // 百度私有化API：未指定参数值时采用配置值；明确指定时则采用参数值
        boolean usePrivateBaiduApi = faceParam.getApiType() == null ? properties.getDefaultPrivateBaiduApi() :
                FaceApiType.PRIVATE_BAIDU_API == faceParam.getApiType();

        Map<String, String> imageData = this.buildImageParamForBaidu(faceParam.getImages()[0], usePrivateBaiduApi);
        imageData.put("face_field", "spoofing,face_type,gender,age");
        imageData.put("liveness_control", "NORMAL");
        imageData.put("max_face_num", "2");
        String params = new JsonHelper().toJson(imageData);

        // 清理大对象内存
        imageData.clear();
        faceParam = null;

        if (usePrivateBaiduApi) {
            return BaiduApiHelper.faceDetectUsePrivateApi(properties, params);
        }
        return BaiduApiHelper.faceDetect(properties, params);
    }

    /**
     * 人脸比对
     */
    @Override
    public FaceResult faceCompare(FaceParam faceParam) {
        if (ArrayUtils.getLength(faceParam.getImages()) < 2) {
            throw new IllegalArgumentException("至少传两个图片参数！");
        }

        if (FaceApiType.FACE_PLUS == faceParam.getApiType()) {
            Map<String, String> params = new HashMap<>();
            this.buildImageParamForFacePlus(params, faceParam.getImages()[0], "1");
            this.buildImageParamForFacePlus(params, faceParam.getImages()[1], "2");

            return FacePlusApiHelper.faceCompare(properties, params);
        }

        // 百度私有化API：未指定参数值时采用配置值；明确指定时则采用参数值
        boolean usePrivateBaiduApi = faceParam.getApiType() == null ? properties.getDefaultPrivateBaiduApi() :
                FaceApiType.PRIVATE_BAIDU_API == faceParam.getApiType();

        List<Map<String, String>> images = new ArrayList<>();
        images.add(this.buildImageParamForBaidu(faceParam.getImages()[0], usePrivateBaiduApi));
        images.add(this.buildImageParamForBaidu(faceParam.getImages()[1], usePrivateBaiduApi));
        String params = new JsonHelper().toJson(images);

        // 百度自定义人脸相似度期望通过的阈值
        Double expectFaceCompareScore = faceParam.getExpectFaceCompareScore() != null ?
                faceParam.getExpectFaceCompareScore() : properties.getBaiduExpectFaceCompareScore();

        // 清理大对象内存
        images.clear();
        faceParam = null;

        if (usePrivateBaiduApi) {
            return BaiduApiHelper.faceCompareUsePrivateApi(properties, params, expectFaceCompareScore);
        }
        return BaiduApiHelper.faceCompare(properties, params, expectFaceCompareScore);
    }

    private Map<String, String> buildImageParamForBaidu(ImageParm image, boolean usePrivateBaiduApi) {
        if (image == null || StringUtils.isBlank(image.value())) {
            throw new IllegalArgumentException("ImageParm must be not empty.");
        }

        // image_type: URL、BASE64、FACE_TOKEN，其中“私有化部署API”不支持 URL 方式
        Map<String, String> imageData = new HashMap<>();
        if (ImageParmType.FILE_URL == image.type()) {
            if (usePrivateBaiduApi) {
                String imageBase64 = CommonUtils.toBase64(image.value());
                imageData.put("image", imageBase64);
                imageData.put("image_type", "BASE64");
            } else {
                imageData.put("image_type", "URL");
                imageData.put("image", image.value());
            }
        } else if (ImageParmType.FACE_TOKEN == image.type()) {
            imageData.put("image_type", "FACE_TOKEN");
            imageData.put("image", image.value());
        } else {
            imageData.put("image_type", "BASE64");
            if (ImageParmType.FILE_PATH == image.type()) {
                File imageFile = new File(image.value());
                String imageBase64 = CommonUtils.toBase64(imageFile);
                imageData.put("image", imageBase64);
            } else {
                imageData.put("image", image.value());
            }
        }
        return imageData;
    }

    private void buildImageParamForFacePlus(Map<String, String> params, ImageParm image, String index) {
        if (image == null || StringUtils.isBlank(image.value())) {
            throw new IllegalArgumentException("ImageParm must be not empty.");
        }

        // 参数：image_url、image_file、image_base64、face_token
        if (ImageParmType.FILE_URL == image.type()) {
            params.put("image_url" + (index != null ? index : ""), image.value());
        } else if (ImageParmType.FACE_TOKEN == image.type()) {
            params.put("face_token" + (index != null ? index : ""), image.value());
        } else {
            if (ImageParmType.FILE_PATH == image.type()) {
                File imageFile = new File(image.value());
                String imageBase64 = CommonUtils.toBase64(imageFile);
                params.put("image_base64" + (index != null ? "_" + index : ""), imageBase64);
            } else {
                params.put("image_base64" + (index != null ? "_" + index : ""), image.value());
            }
        }
    }

    public void setProperties(FaceVerifyProperties properties) {
        this.properties = properties;
    }

}
